import dataclasses
from dataclasses import InitVar
from functools import cached_property
from typing import Callable, Optional, Iterable, Union, Final, Any

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate, ValidationError

from dumbo_asp.primitives.atoms import GroundAtom
from dumbo_asp.primitives.parsers import Parser
from dumbo_asp.primitives.predicates import Predicate

COMPUTE_SUBSTITUTION_UUID: Final = "cace95f4_70b9_44e4_ab5d_9ca8063c798b"


@typeguard.typechecked
@dataclasses.dataclass(frozen=True, order=True)
class Model:
    value: tuple[GroundAtom | int | str, ...]
    is_sorted: bool

    key: InitVar[PrivateKey]
    __key = PrivateKey()

    __compute_substitutions_calls = 0

    class NoModelError(ValueError):
        def __init__(self, *args):
            super().__init__("no stable model", *args)

    class MultipleModelsError(ValueError):
        def __init__(self, *args):
            super().__init__("more than one stable model", *args)

    @staticmethod
    def empty():
        return Model(key=Model.__key, value=(), is_sorted=True)

    @staticmethod
    def of_control(control: clingo.Control, *, sort: bool = True) -> "Model":
        def on_model(model):
            if on_model.cost is not None and on_model.cost <= model.cost:
                on_model.exception = True
            on_model.cost = model.cost
            on_model.res = Model.of_elements(*model.symbols(shown=True), sort=sort)
        on_model.cost = None
        on_model.res = None
        on_model.exception = False

        control.solve(on_model=on_model)
        if on_model.res is None:
            raise Model.NoModelError
        if on_model.exception:
            raise Model.MultipleModelsError
        return on_model.res

    @staticmethod
    def of_program(*args: Any | Iterable[Any], sort: bool = True) -> "Model":
        program = []

        for arg in args:
            if type(arg) is str:
                program.append(str(arg))
            else:
                try:
                    program.extend(str(elem) for elem in arg)
                except TypeError:
                    program.append(str(arg))
        control = clingo.Control()
        control.add('\n'.join(program))
        control.ground([("base", [])])
        return Model.of_control(control, sort=sort)

    @staticmethod
    def of_atoms(
            *args: Union[str, clingo.Symbol, GroundAtom, Iterable[str | clingo.Symbol | GroundAtom]],
            sort: bool = True,
    ) -> "Model":
        res = Model.of_elements(*args, sort=sort)
        validate("only atoms", res.contains_only_ground_atoms, equals=True,
                 help_msg="Use Model.of_elements() to create a model with numbers and strings")
        return res

    @staticmethod
    def of_elements(
            *args: int | str | clingo.Symbol | GroundAtom | Iterable[int | str | clingo.Symbol | GroundAtom],
            sort: bool = True,
    ) -> "Model":
        def build(the_atom):
            if type(the_atom) in [GroundAtom, int]:
                return the_atom
            if type(the_atom) is clingo.Symbol:
                if the_atom.type == clingo.SymbolType.Number:
                    return the_atom.number
                if the_atom.type == clingo.SymbolType.String:
                    return the_atom.string
                return GroundAtom(the_atom)
            if type(the_atom) is str:
                try:
                    return GroundAtom.parse(the_atom)
                except ValidationError:
                    if the_atom[0] == '"' == the_atom[-1]:
                        return Parser.parse_ground_term(the_atom).string
                    return Parser.parse_ground_term(f'"{the_atom}"').string
            return None

        flattened = []
        for element in args:
            built_element = build(element)
            if built_element is not None:
                flattened.append(built_element)
            else:
                for atom in element:
                    built_element = build(atom)
                    validate("is atom", built_element, help_msg=f"Failed to build atom from {element}")
                    flattened.append(built_element)

        model = Model(key=Model.__key, value=tuple(flattened), is_sorted=False)
        return model.sorted if sort else model

    @cached_property
    def sorted(self) -> "Model":
        return self if self.is_sorted else Model(
            key=Model.__key,
            value=
            tuple(sorted(x for x in self if type(x) is int)) +
            tuple(sorted(x for x in self if type(x) is str)) +
            tuple(sorted(x for x in self if type(x) is GroundAtom)),
            is_sorted=True,
        )

    def __post_init__(self, key: PrivateKey):
        self.__key.validate(key)

    def __str__(self):
        return ' '.join(str(x) for x in self.value)

    def __len__(self):
        return len(self.value)

    def __getitem__(self, item):
        return self.value[item]

    def __iter__(self):
        return self.value.__iter__()

    @cached_property
    def contains_only_ground_atoms(self) -> bool:
        return all(type(element) is GroundAtom for element in self)

    @property
    def as_facts(self) -> str:
        def build(element):
            if type(element) is int:
                return f"__number({element})."
            if type(element) is str:
                return f"__string(\"{element}\")."
            return f"{element}."

        return '\n'.join(build(element) for element in self)

    @property
    def as_choice_rules(self) -> str:
        def build(element):
            if type(element) is int:
                return f"{{__number({element})}}."
            if type(element) is str:
                return f"{{__string(\"{element}\")}}."
            return f"{{{element}}}."

        return '\n'.join(build(element) for element in self)

    def drop(self, predicate: Optional[Predicate] = None, numbers: bool = False, strings: bool = False) -> "Model":
        def when(element):
            if type(element) is GroundAtom:
                return predicate is None or not predicate.match(element.predicate)
            if type(element) is int:
                return not numbers
            assert type(element) is str
            return not strings

        return self.filter(when)

    def filter(self, when: Callable[[GroundAtom], bool]) -> "Model":
        return Model(key=self.__key, value=tuple(atom for atom in self if when(atom)), is_sorted=self.is_sorted)

    def map(self, fun: Callable[[GroundAtom], GroundAtom]) -> 'Model':
        return Model(key=self.__key, value=tuple(sorted(fun(atom) for atom in self)), is_sorted=self.is_sorted)

    def rename(self, predicate: Predicate, new_name: Predicate) -> "Model":
        validate("same arity", predicate.arity == new_name.arity, equals=True,
                 help_msg="Predicates must have the same arity")
        return self.map(lambda atom: atom if not predicate.match(atom.predicate) else GroundAtom(
            clingo.Function(new_name.name, atom.arguments)
        ))

    def substitute(self, predicate: Predicate, argument: int, term: clingo.Symbol) -> "Model":
        validate("argument", argument, min_value=1, max_value=predicate.arity, help_msg="Arguments are indexed from 1")

        def mapping(atom: GroundAtom) -> GroundAtom:
            if not predicate.match(atom.predicate):
                return atom
            return GroundAtom(clingo.Function(
                atom.predicate_name,
                [arg if index != argument else term for index, arg in enumerate(atom.arguments, start=1)]
            ))

        return self.map(mapping)

    def project(self, predicate: Predicate, argument: int) -> "Model":
        validate("argument", argument, min_value=1, max_value=predicate.arity, help_msg="Arguments are indexed from 1")

        def mapping(atom: GroundAtom) -> GroundAtom:
            if not predicate.match(atom.predicate):
                return atom
            return GroundAtom(clingo.Function(
                atom.predicate_name,
                [arg for index, arg in enumerate(atom.arguments, start=1) if index != argument]
            ))

        return self.map(mapping)

    @property
    def block_up(self) -> str:
        return ":- " + ", ".join([f"{atom}" for atom in self]) + '.'

    @cached_property
    def __compute_substitutions_control(self):
        program = self.as_choice_rules
        control = clingo.Control()
        control.add(program)
        control.ground([("base", [])])
        return control

    def compute_substitutions(self, *, arguments: str, number_of_arguments: int,
                              conjunctive_query: str) -> tuple[list[clingo.Symbol], ...]:
        Model.__compute_substitutions_calls += 1
        predicate: Final = f"__query_{COMPUTE_SUBSTITUTION_UUID}_{Model.__compute_substitutions_calls}__"
        self.__compute_substitutions_control.add(predicate, [], f"{predicate}({arguments}) :- {conjunctive_query}.")
        self.__compute_substitutions_control.ground([(predicate, [])])
        return tuple(
            atom.symbol.arguments
            for atom in self.__compute_substitutions_control.symbolic_atoms.by_signature(predicate, number_of_arguments)
        )



import dataclasses
import math
from dataclasses import InitVar

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate


@typeguard.typechecked
class Parser:
    @dataclasses.dataclass(frozen=True)
    class Error(ValueError):
        parsed_string: str
        line: int
        begin: int
        end: int
        message: str

        key: InitVar[PrivateKey]
        __key = PrivateKey()

        def __post_init__(self, key: PrivateKey):
            self.__key.validate(key)

        @staticmethod
        def parse(error: str, parsed_string: str) -> "Parser.Error":
            parts = error.split(':', maxsplit=3)
            validate("prefix", parts[0], equals="<string>", help_msg="Unexpected source")
            validate("error", parts[3].startswith(" error: "), equals=True, help_msg="Unexpected error")
            begin, end = Parser.parse_range(parts[2])
            return Parser.Error(
                parsed_string=parsed_string,
                line=int(parts[1]),
                begin=begin,
                end=end,
                message=parts[3][len(" error: "):],
                key=Parser.Error.__key,
            )

        def drop(self, *, first: int = 0, last: int = 0) -> "Parser.Error":
            validate("one line", self.line, equals=1, help_msg="Can drop only from one line parsing")
            return Parser.Error(
                parsed_string=self.parsed_string[first:len(self.parsed_string) - last],
                line=self.line,
                begin=self.begin - first,
                end=self.end - first,
                message=self.message,
                key=Parser.Error.__key,
            )

        def __str__(self):
            lines = self.parsed_string.split('\n')
            width = math.floor(math.log10(len(lines))) + 1
            res = [f"Parsing error in line {self.line}, columns {self.begin}-{self.end}"]
            for line_index, the_line in enumerate(lines, start=1):
                res.append(f"{str(line_index).zfill(width)}| {the_line}")
                if line_index == self.line:
                    res.append('>' * width + '| ' + ' ' * (self.begin - 1) + '^' * (self.end - self.begin + 1))
            res.append(f"error: {self.message}")
            return '\n'.join(res)

    @staticmethod
    def parse_range(string: str) -> tuple[int, int]:
        parts = string.split('-', maxsplit=1)
        if len(parts) == 1:
            return int(parts[0]), int(parts[0])
        return int(parts[0]), int(parts[1])

    @staticmethod
    def parse_ground_term(string: str) -> clingo.Symbol:
        try:
            return clingo.parse_term(string)
        except RuntimeError as err:
            raise Parser.Error.parse(str(err), string)

    @staticmethod
    def parse_program(string: str) -> list[clingo.ast.AST]:
        def callback(ast):
            callback.res.append(ast)
        callback.res = []

        messages = []
        try:
            clingo.ast.parse_string(string, callback, logger=lambda code, message: messages.append((code, message)))
            validate("nonempty res", callback.res, min_len=1)
            validate("base program", callback.res[0].ast_type == clingo.ast.ASTType.Program and
                     callback.res[0].name == "base" and len(callback.res[0].parameters) == 0, equals=True)
            res = [x for x in callback.res[1:] if x.ast_type != clingo.ast.ASTType.Comment]
            validate("only rules", [x for x in res if x.ast_type != clingo.ast.ASTType.Rule], empty=True)
            return res
        except RuntimeError:
            errors = [message[1] for message in messages if message[0] == clingo.MessageCode.RuntimeError]
            validate("errors", messages, length=1)
            raise Parser.Error.parse(errors[0], string)




import dataclasses
from dataclasses import InitVar
from pathlib import Path

import clingo
import clingo.ast
import typeguard
from dumbo_utils.primitives import PrivateKey
from dumbo_utils.validation import validate

from dumbo_asp import utils
from dumbo_asp.primitives.predicates import Predicate
from dumbo_asp.primitives.programs import SymbolicProgram


@typeguard.typechecked
@dataclasses.dataclass(frozen=True)
class Template:
    @dataclasses.dataclass(frozen=True)
    class Name:
        value: str
        key: InitVar[PrivateKey]
        __key = PrivateKey()

        def __post_init__(self, key: PrivateKey):
            self.__key.validate(key)

        @staticmethod
        def parse(name: str) -> "Template.Name":
            term = clingo.String(name)
            return Template.Name(
                value=term.string,
                key=Template.Name.__key,
            )

        def __str__(self):
            return self.value

    name: "Template.Name"
    program: SymbolicProgram
    __static_uuid: str = dataclasses.field(default_factory=lambda: utils.uuid(), init=False)

    __core_templates = None
    __core_templates_directory = Path(__file__).parent.parent / "templates"
    __core_templates_files = [f"{name}.template.asp" for name in [
        "dumbo",
        "binary_relations",
        "sets",
        "graphs",
        "grids",
    ]]

    @staticmethod
    def __init_core_templates():
        if Template.__core_templates is not None:
            return
        Template.__core_templates = {}

        def register(template: str):
            name, program = template.strip().split('\n', maxsplit=1)
            name = f"@dumbo/{name}"
            assert name not in Template.__core_templates
            Template.__core_templates[name] = Template(
                Template.Name.parse(name),
                Template.expand_program(SymbolicProgram.parse(program.strip())),
            )

        def register_all(templates: str, *, sep="----"):
            for template in templates.strip().split(sep):
                if template:
                    lines = [line[4:] if index > 0 and len(line) >= 4 else line
                             for index, line in enumerate(template.strip().split('\n'))]
                    register('\n'.join(lines))

        for arity in range(10):
            terms = ','.join('X' + str(i) for i in range(arity))
            register(f"""
exact copy (arity {arity})
output({terms}) :- input({terms}).
:- output({terms}), not input({terms}).
            """)
            if arity > 0:
                register(f"collect arguments (arity {arity})\n" +
                         '\n'.join(f"output(X{index}) :- input({terms})." for index in range(arity)))
                for other_arity in range(arity):
                    register(f"collect argument {other_arity + 1} of {arity}\n" +
                             f"output(X{other_arity}) :- input({terms}).")

        for file in Template.__core_templates_files:
            with open(Template.__core_templates_directory / file) as templates_file:
                Template.expand_program(SymbolicProgram.parse(templates_file.read()), register_templates=True)

    @staticmethod
    def core_template(name: str) -> "Template":
        Template.__init_core_templates()
        return Template.__core_templates[name]

    @staticmethod
    def is_core_template(name: str) -> bool:
        Template.__init_core_templates()
        return name in Template.__core_templates

    @staticmethod
    def core_templates() -> int:
        Template.__init_core_templates()
        return len(Template.__core_templates)

    @staticmethod
    def core_template_names() -> tuple[str, ...]:
        Template.__init_core_templates()
        return tuple(Template.__core_templates.keys())

    @staticmethod
    def core_templates_as_parsable_string() -> str:
        Template.__init_core_templates()
        res = []
        for key, value in Template.__core_templates.items():
            res.append(str(value))
        return '\n'.join(res)

    @staticmethod
    def expand_program(program: SymbolicProgram, *, limit: int = 100_000, register_templates: bool = False,
                       trace: bool = False) -> SymbolicProgram:
        Template.__init_core_templates()
        templates = {}
        template_under_read = None
        res = []
        for rule in program:
            validate("avoid blow up", len(res) + (len(template_under_read[1]) if template_under_read else 0),
                     max_value=limit,
                     help_msg=f"The expansion takes more than {limit} rules. "
                              f"If you trust the code, try again by increasing the limit.")
            if rule.disabled or not rule.is_normal_rule:
                if template_under_read is not None:
                    template_under_read[1].append(rule)
                else:
                    res.append(rule)
            elif rule.head_atom.predicate_name == "__template__":
                validate("empty body", rule.is_fact, equals=True)
                validate("arity 1", rule.head_atom.predicate_arity, equals=1)
                validate("arg#0", rule.head_atom.arguments[0].is_string(), equals=True)
                validate("no nesting", template_under_read is None, equals=True)
                validate("not a core template", Template.is_core_template(rule.head_atom.arguments[0].string_value()), equals=False)
                validate("not seen", rule.head_atom.arguments[0].string_value() not in templates, equals=True)
                template_under_read = (rule.head_atom.arguments[0].string_value(), [])
            elif rule.head_atom.predicate_name == "__end__":
                validate("empty body", rule.is_fact, equals=True)
                validate("arity 0", rule.head_atom.predicate_arity, equals=0)
                validate("not in a template", template_under_read)
                if trace:
                    template_under_read[1].append(rule.disable())
                the_template = Template(name=Template.Name.parse(template_under_read[0]),
                                        program=SymbolicProgram.of(template_under_read[1]))
                if register_templates:
                    Template.__core_templates[template_under_read[0]] = the_template
                else:
                    templates[template_under_read[0]] = the_template
                template_under_read = None
            elif rule.head_atom.predicate_name == "__apply_template__":
                validate("empty body", rule.is_fact, equals=True)
                validate("arity >= 1", rule.head_atom.predicate_arity, min_value=1)
                validate("arg#0", rule.head_atom.arguments[0].is_string(), equals=True)
                template_name = rule.head_atom.arguments[0].string_value()
                if Template.is_core_template(template_name):
                    template = Template.core_template(template_name)
                else:
                    validate("known template", template_name in templates, equals=True,
                             help_msg=f"Unknown template: {template_name}")
                    template = templates[template_name]
                mapping = {}
                for argument in rule.head_atom.arguments[1:]:
                    validate("mapping args", argument.is_function(), equals=True)
                    validate("mapping args", argument.function_name, equals='')
                    validate("mapping args", argument.function_arity, equals=2)
                    validate("mapping args", argument.arguments[0].is_function(), equals=True)
                    validate("mapping args", argument.arguments[0].function_arity, max_value=1)
                    validate("mapping args", argument.arguments[1].is_function(), equals=True)
                    validate("mapping args", argument.arguments[1].function_arity, equals=0)
                    key = str(argument.arguments[0].function_name)
                    if argument.arguments[0].function_arity == 1:
                        validate("mapping args", argument.arguments[0].arguments[0].is_int(), equals=True)
                        key += "/" + str(argument.arguments[0].arguments[0])
                    mapping[key] = Predicate.parse(argument.arguments[1].function_name)
                if template_under_read is None:
                    if trace:
                        res.append(rule.disable())
                    res.extend(r for r in template.instantiate(**mapping))
                else:
                    if trace:
                        template_under_read[1].append(rule.disable())
                    template_under_read[1].extend(r for r in template.instantiate(**mapping))
            elif template_under_read is not None:
                template_under_read[1].append(rule)
            else:
                res.append(rule)

        return SymbolicProgram.of(res)

    def __str__(self):
        return f"""__template__("{self.name}").\n{self.program}\n__end__."""

    def __repr__(self):
        return f"""Template(name="{self.name}", program={self.program})"""

    def instantiate(self, **kwargs: Predicate) -> SymbolicProgram:
        Template.__init_core_templates()
        for arg in kwargs:
            validate("kwargs", arg.startswith('__'), equals=False,
                     help_msg="Local predicates cannot be renamed externally.")
        static_uuid = self.__static_uuid
        local_uuid = utils.uuid()
        mapping = {**kwargs}
        for predicate in self.program.predicates:
            if not predicate.name.endswith('__'):
                if predicate.name.startswith('__static_'):
                    mapping[predicate.name] = Predicate.parse(f"{predicate.name[1:]}_{static_uuid}")
                elif predicate.name.startswith('__'):
                    mapping[predicate.name] = Predicate.parse(f"{predicate.name}_{local_uuid}")
        return self.program.apply_predicate_renaming(**mapping)

    def predicates(self) -> tuple[Predicate, ...]:
        return tuple(predicate for predicate in self.program.predicates if predicate.name.endswith('__') or not predicate.name.startswith('__'))
